;
; FSK modem for HF RTTY
;
; See readme.txt file in this distribution for copyright notice, setup
; instructions and command summary.
;
; This program implements a 45/50/75/100-baud FSK modem for HF Baudot
; asynchronous signalling. It uses Butterworth and matched filters,
; soft-decision detection, maximum a-priory (MAP) estimation and Viterbi
; decoding. In case of continuous signalling, it can operate in a
; synchronous mode using a phase-lock loop to flywheel through momentary
; carrier fades. The predetection design and many of the operator
; features were inspired by the Dovetron TU of long ago, but implemented
; in DSP, rather than analog form. The postdetection design, including
; MAP estimation, Viterbi decoding and synchronous mode were inspired by
; a LSI-11 program also of long ago. If this thing can't pull RTTY
; signals from the muck, nothing can.
;
; Include files
;
#INCLUDE "regs.inc"		; TMC320 registers
#INCLUDE "ports.inc"		; DSP93 ports
#INCLUDE "serial.inc"		; 16550 UART

;
; Program memory map
;
; External	1000-7fff	program
;
; Data memory map
;
; Block B2	0060-007f	interrupt variables and stack
; Block B0	0200-02ff	filter coefficients
; Block B1	0300-03ff	program variables and filter delay lines
; External	0400-7fff	program variables
;
; Auxiliary register usage
;
; AR0	utility compare and temporary register
; AR1	utility pointer
; AR2	utility counter
; AR3	spare
; AR4	spare
; AR5	interrupt stack pointer
; AR6	interrupt buffer pointer
; AR7	interrupt buffer pointer
;
; Status bits (STATUS)
;
SYNTST	.EQU	0		; sync mode
AUTTST	.EQU	1		; pulse gate
IIRTST	.EQU	2		; use Butterworth (IIR) filter
IIRBIT	.EQU	1<<IIRTST	; shifted IIRTST
WIDTST	.EQU	3		; use wide filters
WIDBIT	.EQU	1<<WIDTST	; shifted WIDTST
CARTST	.EQU	4		; signal gate
ENBTST 	.EQU	5		; sync mode enable
PGATE	.EQU	6		; pulse gate enable
RXREV	.EQU	7		; reverse shift
;
; AIO Configuration constants: HPF in, sync, 3-V half scale, sin x/x.
; RA is chosen to get the SCF as close to 288 kHz as possible; RB is
; chosen to get the sample frequency as close to 8.0 kHz as possible.
; This gives a 3-dB bandwidth of about 100-3700 Hz at the AIO input.
;
CFGA	.EQU	CMDA_VAL(17)	; AIO load RA=TA=17 for SCF 294.1 kHz
CFGB	.EQU	CMDB_VAL(37)	; AIO load RB=TB=37 for 7.949 Hz
AIO_GAIN .EQU	AIO_GAIND60|AIO_GAIND70 ; AIO gain
CFGC	.EQU	AIO_CONFIG|AIO_INSADHPF|AIO_SYNC|AIO_GAIN|AIO_INSSINX
;
; Tone frequency definitions. Note input uses inverted tones, (mark 2295
; Hz, space lower), while regenerator output uses upright tones
; expected by most TNCs. 
;
DEGREE	.EQU	360/3		; sin table entries (3-degree steps)
PI2	.EQU	DEGREE<<8	; radix for output phase wrap
SFREQ	.EQU	7949		; sample frequency (5000000 / (17 * 37))
INCR	.EQU	126		; sample interval (us) (1000000 / SFREQ)
TUNEX	.EQU	39		; 10 Hz (10 * 120 * 256 / SFREQ)
;
; Character format definitions. Note that the number of bits here is one
; more than the actual number, in order that the mark signal preceding
; the character can be evaluated.
;
BINSIZ	.EQU	8		; samples per baud interval
CHRSIZ	.EQU	8		; baud intervals per character + 1
SRVSIZ	.EQU	CHRSIZ*BINSIZ	; size of survivor table
;
; Miscellaneous definitions
;
LIMIT	.EQU	10000		; limiter threshold (LED 1)
ALIMIT	.EQU	16000		; autostart preset threshold
CLIMIT	.EQU	5000		; carrier preset threshold
FIRLEN	.EQU	51		; length of FIR lpf delay line
IIRLEN	.EQU	3*4		; length of IIR lpf delay lines (4)
BUFLEN	.EQU	50		; length of interrupt buffer
SYNJTR	.EQU	BINSIZ/2	; sync counter max jitter
FRMSIZ	.EQU	SRVSIZ-BINSIZ	; sync counter radix (samples / char)
ACONST	.EQU	16-3		; scaled autostart time constant
CCONST	.EQU	16-5		; scaled carrier/noise time constant
TCONST	.EQU	16-4		; scaled acquistion pll time constant
SCONST	.EQU	16-4		; scaled tracking pll time constant
PGAIN	.EQU	4+5		; scaled pll gain

;
; Interesting block addresses, memory page pointers and offsets
;
B_PAGE	.EQU	0h		; base page pointer
D_PAGE	.EQU	6h		; data page pointer
D_ADDR	.EQU	D_PAGE<<7	; data page address
U_PAGE	.EQU	8h		; user page pointer
U_ADDR	.EQU	U_PAGE<<7	; user page address
B0P	.EQU	0FF00h		; block B0 program address
B0D	.EQU	200h		; block B0 data address
B1D	.EQU	300h		; block B1 data address
BPFCOF	.EQU	B0P		; FIR coefficients address (FIRLEN)
LPFCOF	.EQU	B0P+FIRLEN	; IIR coefficients address (IIRLEN)

;
; Following are variables accessed by either absolute or indirect
; addressing modes.
;
; Program variables in block B1 (D_PAGE) (cleared at reset)
;
VAREND	.EQU	D_ADDR+30h	; leave room for 48 variables
;
; Delay lines. The MACD instruction needs these in internal memory.
;
DELBPF	.EQU	VAREND		; FIR bandpass filter delay line
DELIIR	.EQU	DELBPF+FIRLEN+1	; IIR lowpass filter delay line
DELEND	.EQU	DELIIR+IIRLEN	; end of delay lines
;
; The survivor table contains a shift-register survivor for 8 samples of
; each bit and 8 bits of the character. It is used to construct the
; maximum and minimum amplitudes, as well as weighted Viterbi distances
; for each candidate survivor. The entries in the space sample table
; correspond to those in the survivor table. These are used to estimate
; the noise floor for the carrier signal.
;
SRVBGN	.EQU	DELEND		; beginning of survivor table
SRVEND	.EQU	SRVBGN+SRVSIZ	; end of survivor table
;
; These structures are used by both the IIR and MF filters. The mark and
; space filter center frequencies are determined by the phase increment.
; The mark filter is fixed; the space filter is tunable.
;
; Structure format
; 0	phase increment
; 1	current phase
; 2-3	I channel output
; 4-5	Q channel output
; 6-7	sum of squares
; 8-9	detector output
;
MSTRUCT	.EQU	SRVEND		; mark filter structure
SSTRUCT	.EQU	MSTRUCT+10	; space filter structure
;
; The interrupt buffer is a circular structure indexed by AR6 and AR7.
; AR6 is used only by interrupt routines. The receive interrupt routine
; stores the sample at AR6 and then increments AR6. The transmit
; interrupt routine loads the sample at AR6, which is the oldest in the
; circular buffer. The main program loops until AR7 is not equal to AR6,
; then loads the sample found at AR7, which is the oldest sample not yet
; processed. The program loads the output at AR7 and then increments
; AR7. Since input and output interrupts are synchronous, no data are
; lost or duplicated.
;
INTBUF	.EQU	SSTRUCT+10	; interrupt buffer
D_END	.EQU	INTBUF+BUFLEN	; must be less than 400h
;
; Program variables in external memory (U_PAGE) (cleared at reset). We
; protect one location at the beginning for I/O status.
;
USRBGN	.EQU	U_ADDR+1	; beginning of clear region
;
; The matched filter. It can grow up to 860 words, depending on baud
; interval. In the predetection case, it contains four samples for each
; cycle; in the baseband case, it contains two samples interleaved with
; two dummy words for a total of four.
;
DELMFR	.EQU	U_ADDR+100h	; may be up to 700 words
MRKBGN	.EQU	DELMFR+1000	; beginning of space sample table
MRKEND	.EQU	MRKBGN+SRVSIZ	; end of space sample table
SPCBGN	.EQU	MRKEND		; beginning of space sample table
SPCEND	.EQU	SPCBGN+SRVSIZ	; end of space sample table
;
USREND	.EQU	SPCEND		; end of clear region 

;
; Following are variables accessed by direct addressing mode. These are
; offsets relative the the data page pointer.
;
: Program variable offsets in block B2 (B_PAGE) initialized at startup
;
TNC_BUF	.EQU	60h		; TNC_OUTPUT port status
PDC_BUF	.EQU	61h		; RADIO_CTRL port status
;
; Program variable offsets in block B2 (B_PAGE) cleared at reset
;
UARTI	.EQU	62h		; UART input buffer
UARTO	.EQU	63h		; UART output buffer
XTMP	.EQU	64h		; temp
STACK	.EQU	80h		; interrupt stack (grows down)
;
; Program variable offsets in block B1 (D_PAGE) cleared at reset
;
; Analog signals and pointers
;
BUFI	.EQU	00h		; AIO input buffer
SIGNAL	.EQU	01h		; filtered/limited input signal
MARK	.EQU	02h		; mark channel output
SPACE	.EQU	03h		; space channel output
ACCUM	.EQU	04h		; detector output
POSMAX	.EQU	05h		; max positive signal
NEGMAX	.EQU	06h		; max negative signal
MFPTR	.EQU	07h		; matched filter pointer
MSPTR	.EQU	08h		; mark memory pointer
CARDET	.EQU	09h		; signal + noise estimate
NOISE	.EQU	0Ah		; noise estimate
DIST	.EQU	0Bh		; max survivor distance
;
; Regenerator variables
;
BITCTR	.EQU	0Ch		; survivor bit counter
SRVCTR	.EQU	0Dh		; survivor strobe counter
KEY	.EQU	0Eh		; output key line
;
; Tuning parameters
;
RANGE	.EQU	12h		; baud interval counter radix
COUNT	.EQU	13h		; baud interval counter
MFSIZE	.EQU	14h		; matched filter size
MFGAIN	.EQU	15h		; matched filter gain
GAIN	.EQU	16h		; filter gain
LGAIN	.EQU	17h		; limiter gain (shift)
;
; Signal quality variables
;
STATUS	.EQU	18h		; status bits
SYNCTR	.EQU	19h		; sync counter
SYNDEL	.EQU	1Ah		; frame interval
XPHASE	.EQU	1Bh		; phase error signal
AUTOS	.EQU	1Ch		; autostart signal
CTHRESH	.EQU	1Dh		; carrier threshold
ATHRESH	.EQU	1Eh		; autostart threshold
;
; Local temporaries
;
TMP	.EQU	1Fh		; temporary mongrel
STS	.EQU	20h		; temporary data
SPARE1	.EQU	21h		; spare
;
; Output variables
;
OUTSW	.EQU	22h		; display pointer
OMARK	.EQU	23h		; AFSK mark phase increment
OSPACE	.EQU	24h		; AFSK space phase increment
OPHASE	.EQU	25h		; AFSK output phase
;
; Command interpreter variables
;
UTMP1	.EQU	26h		; user temp 1
UTMP2	.EQU	27h		; user temp 2
UCHAR	.EQU	28h		; character temp
UPTR	.EQU	29h		; pointer temp
UADR	.EQU	2Ah		; address temp
UCMD	.EQU	2Bh		; command table pointer
UMOD	.EQU	2Ch		; saved command table pointer
;
; Program variable offsets in external memory (U_PAGE) initialized at
; startup
;
PIO_BUF	.EQU	00h		; RADIO_GAIN port status

;
;  Program transfer vector
;
	.ORG	1000h		; monitor origin
	B       TINT		; timer interrupt
	B       RINT		; AIO receive interrupt
XIN10	B       XINT		; AIO transmit interrupt
	B       IINT		; TRAP instruction interrupt
	B       GO		; program starting address

;
; Initialize processor configuration. These set the options as does the
; CPU upon first coming up; they are here in case the monitor screws up.
;
GO	CNFP			; configure block B0 as program
	SSXM			; set sign extension mode
	SOVM			; set overflow mode
	SPM	1		; set P register output shift for q15
	FORT	0		; set AIO 16-bit mode
	SFSM			; set FSM (1) frame sync mode
	RTXM			; set TXM (0) transmit mode
;
; Initialize ports and program variables
;
	LDPK	U_PAGE
	LALK	AIO_RESET+GAIN4+REC_IN_1 ; reset AIO
	SACL	PIO_BUF
	OUT	PIO_BUF,RADIO_GAIN
	CALL    WAIT4
	LALK    LED_202+GAIN4+REC_IN_1 ; disable AIO; LED 2 on
	SACL    PIO_BUF
	OUT     PIO_BUF,RADIO_GAIN 
	CALL    WAIT4
	LALK    AIO_ENABLE+GAIN4+REC_IN_1 ;enable AIO
	SACL    PIO_BUF        
	OUT     PIO_BUF,RADIO_GAIN
	CALL    WAIT4
	LDPK	B_PAGE
	ZAC			; initialize TNC interface
	SACL	PIO_BUF        
	OUT	PIO_BUF,TNC_OUTPUT
	LALK    FREQ_DN_R2+FREQ_UP_R2+FREQ_DN_R1+FREQ_UP_R1 ; initialize  radio interface
	SACL    PDC_BUF
	OUT     PDC_BUF,RADIO_CTRL
	RPTK	7-1		; flush stack
	POP
	LRLK	AR5,STACK	; initialize stack pointer
	LRLK	AR6,INTBUF	; initialize interrupt buffer pointers
	LRLK	AR7,INTBUF
;
; Set up for AIO initialization. Keep this a deep dark secret and don't
; tell anybody the transmit interrupt routines are switched on-the-wing
; for configuration. This is because the AIO rascal has unreal latency
; requirements only during initialization. There seems to be no way for
; an interrupt routine to satisfy these without appearing non-
; transparent for general purpose processing.
;
	DINT			; turn out the lights
	LALK	XINT2
	SACL	XTMP
	LALK	XIN10+1		; clobber transmit vector
	TBLW	XTMP
	LARP	AR7		; copy AIO initialization vector
	RPTK	7-1
	BLKP	INITAB,*+
	LACK	20h		; enable AIO tx interrupt
	SACL	IMR
	ZAC			; flush AIO tx buffer
	SACL	DXR
	EINT			; turn on the lights
	LARP	AR6		; load initialization word
INI10	IDLE			; wait for interrupt and AC
	BNZ	INI10		; branch if more
	DINT			; done. darkness again
	LALK	XINT		; unclobber transmit vector
	SACL	XTMP
	LALK	XIN10+1
	TBLW	XTMP
	LACK	30h		; enable AIO rx and tx interrupts
	SACL	IMR
	LDPK	D_PAGE
	CALL	RESET		; initialize program variables
	EINT			; enable interrupts
;
; Main loop. Loop looking for a sample in the input buffer. When found,
; process it and develop the output signal. All this stuff runs with
; interrupts enabled, except for the part of the command-input routine.
;
; Note: This is the only place where AR6 is compared to AR7. The CMPR
; insruction represents the atomic interaction in order to determine
; whether data are in the buffer or not; thus, there are no races. Even
; if there were, unread data would get caught on the next cycle through
; the loop.
;
LOOP	LARP	AR7		; AR7 is interrupt buffer pointer
	SAR	AR6,TMP		; are there data in the buffer
	LAR	AR0,TMP
	CMPR
	BBNZ	MAN10		; branch if no
	CALL	DSPIN		; predetection processing
	CALL	DSPOUT		; postdetection processing
	CALL	OUTX		; produce the output signal
	B	LOOP
;
MAN10	LALK	LED_202		; LED 2 on
	CALL    LEDON
	CALL	LIGHTS		; blink the control panel
	CALL	GETCH		; process keyboard commands
	LALK	LED_202		; LED 2 off
	CALL    LEDOFF
	B       LOOP

;
; AIO receive interrupt. This is designed for minimal intrusion on
; variables used by interruptable code. AIO rx interrupt is enabled when
; tx initialization is complete.
;
; Note: This is the only place where AR6 is incremented. Immediately
; after increment, AR6 is checked for overflow and wrapped.
;
; Variables
; AR6		interrupt buffer pointer
; DRR		AIO input buffer
;
RINT	LARP	AR5		; save context
	MAR	*-
	SST1	*-
	SST	*-
	SAR	AR0,*
	LDPK	B_PAGE
	LARP	AR6		; AR6 is interrupt buffer pointer
	LAR	AR0,DRR		; read AIO input buffer
	SAR	AR0,*+
	LRLK	AR0,INTBUF+BUFLEN ; is pointer at buffer end
	CMPR	
	BBZ	RIN10		; branch if no
	LRLK	AR6,INTBUF	; yes. reset pointer
RIN10	LARP	AR5
	CALL	DUMP		; dump status
	LAR	AR0,*+		; restore context
	LST	*+
	LST1	*+
	EINT			; enable interrupts
	RET

;
; AIO transmit interrupt used in normal processing
;
; Variables
; AR6		interrupt buffer pointer
; DXR		AIO output buffer
;
XINT	LARP	AR5		; save context
	MAR	*-
	SST1	*-
	SST	*-
	SAR	AR0,*
	LDPK	B_PAGE
	LARP	AR6		; AR6 is interrupt buffer pointer
	LAR	AR0,*		; write AIO output buffer
	SAR	AR0,DXR
	LARP	AR5
	CALL	DUMP		; dump status
	LAR	AR0,*+		; restore context
	LST	*+
	LST1	*+
	EINT			; enable interrupts
	RET

;
; AIO transmit interrupt routine used only during AIO configuration
;
XINT2	LAC	*+		; this has to be quick and ugly
	SACL	DXR
	EINT
	RET

;
; Trap interrupt (for debug)
;
IINT	DINT			; avoid interrupts here
	LARP	AR5		; save context
	MAR	*-
	SST1	*-
	SST	*-
	SAR	AR0,*
	LDPK	B_PAGE
	CALL	DUMP		; dump status
	LAR	AR0,*+		; restore context
	LST	*+
	LST1	*+
	EINT			; enable interrupts
	RET

;
; Routine to dump trace data for debug. This is done on every I/O and
; TRAP interrupt. TRAP instructions can be sprinkled in the code in
; order to watch critical sections. At this point, SS1, SS0 and AR0 are
; already on the stack (AR5) and the data page pointer set to B_PAGE
; (0). The remaining registers and top two words on the internal CPU
; stack are saved.
;
DUMP	SAR	AR5,XTMP	; save stack pointer
	MAR	*-
	SAR	AR1,*-		; save aux registers AR1-AR7
	SAR	AR2,*-
	SAR	AR3,*-
	SAR	AR4,*-
	SAR	AR5,*-
	SAR	AR6,*-
	SAR	AR7,*-
	POPD	*-		; save calling address
	POPD	*		; save interrupting address
	PSHD	*+
	PSHD	*-
	LAR	AR5,XTMP	; restore stack pointer
	RET

;
; Timer interrupt (not used)
;
TINT	EINT			; enable interrupts
	RET

;
; Predetection processing
;
; Variables
; AR7		interrupt buffer pointer
; BUFI		AIO input buffer
; SIGNAL	filtered/limited input signal
; MARK		mark channel output
; SPACE		space channel output
; DELBPF	FIR bandpass filter delay line
; DELIIR	IIR lowpass filter delay line
; MSTRUCT	mark filter structure
; SSTRUCT	space filter structure
;
; Limit signals with energy above an arbitrary threshold (LIMIT). This
; helps to reduce errors due to noise spikes and static crashes, etc.
; The threshold is chosen rather arbitrary at 10dB below overload. LED 1
; is lit when in limiting condition.
;
DSPIN	TRAP
	LARP	AR7		; AR7 is interrupt buffer pointer
	LAC	*,2		; load sample in q15 format
	SACL	BUFI
	ABS			; is signal above threshold
	SBLK	LIMIT
	BLZ	DSP10		; branch if no
	LALK	LED_201		; yes. LED 1 on
	CALL	LEDON
	B	DSP11
;
DSP10	LALK	LED_201		; signal below threshold; LED 1 off
	CALL	LEDOFF
DSP11	LAC	BUFI		; limit signal at max
	SBLK	LIMIT
	BLZ	DSP12
	ZAC
DSP12	ADLK	2*LIMIT		; limit signal at min
	BGZ	DSP13
	ZAC
DSP13	SBLK	LIMIT
	SACL	SIGNAL
;
; Filter the clipped signal using a FIR filter. The filter bandwidth is
; determined by the coefficients loaded in block B0P.
;
	LARP	AR1		; AR1 is delay line pointer
	LRLK	AR1,DELBPF
	LAC	SIGNAL
	SACL	*
	ADRK	FIRLEN-1	; point to end of delay line
	MPYK	0		; initialize sum and product registers
	ZAC
	RPTK	FIRLEN-1	; multiply and accumulate
	MACD	BPFCOF,*-
DSP21	APAC			; include last product
	SACH	SIGNAL
;
; Limit signal at selected gain 0-30dB. Under good conditions, up to 30-
; dB limiting can be used with IIR filters for ease of tuning. When
; things get tough, use less limiting and matched filters.
; 
	LT	LGAIN		; get limiter gain
	LACT	SIGNAL		; limit signal at max
	SBLK	LIMIT
	BLZ	DSP30
	ZAC
DSP30	ADLK	2*LIMIT		; limit signal at min
	BGZ	DSP31
	ZAC
DSP31	SBLK	LIMIT
	SACL	SIGNAL
DSP32
;
; The receiver is centered first to the mark frequency (2295 Hz). The
; space frequency (variable 2125, 1870, 1445 Hz) is selected with the m
; command, then the t+ and t- commands used as necessary to center the
; space signal in the passband.
;
	BIT	STATUS,IIRTST	; select filter
	BBZ	DSP40		; branch if MF filter
	LRLK	AR2,DELIIR	; IIR filter; process mark signal
	LRLK	AR1,MSTRUCT
	CALL	IFILT
	SACL	MARK
	LRLK	AR1,SSTRUCT	; process space signal
	CALL	IFILT
	SACL	SPACE
	B	DSP41
;
DSP40	LRLK	AR1,MSTRUCT	; MF filter; process mark signal
	CALL	MFILT
	SACL	MARK
	LRLK	AR1,SSTRUCT	; process space signal
	CALL	MFILT
	SACL	SPACE
DSP41	LAC	MFPTR		; adjust filter pointer
	SBLK	DELMFR
	SUB	MFSIZE
	BLZ	DSP42
	LALK	DELMFR
	SACL	MFPTR
DSP42	RET

;
; Baseband processing
;
; ACCUM		detector output
; POSMAX	max positive signal
; NEGMAX	max negative signal
; CARDET	carrier estimate
; NOISE		noise estimate
; DIST		max survivor distance
; BITCTR	survivor bit counter
; SRVCTR	survivor strobe counter
; KEY		output key line
; RANGE		baud interval counter radix
; COUNT		baud interval counter
; STATUS	status bits
; SYNCTR	sync counter
; SYNDEL	frame interval
; XPHASE	phase error signal
; CARDET	carrier signal
; AUTOS		autostart signal
; SRVBGN	survivor table
; MRKBGN	mark sample memory
; SPCBGN	space sample memory
; CTHRESH	carrier threshold
; ATHRESH	autostart threshold
;
; Decimate at 8 times the baud rate, i.e., 8 samples per bit interval.
; At each sample determine the differential signal and the common-mode
; signal. Then compute the max and min of each survivor and the Viterbi
; distances. Finally, select the maximum survivor and trip the character
; clutch.
;
DSPOUT	LAC	COUNT		; is it time for the next update
	SUBK	INCR
	SACL	COUNT
	BGEZ	CYC99		; branch if no
	ADD	RANGE		; yes. restore decimation counter
	SACL	COUNT
	LAC	SYNCTR		; increment sync counter
	ADLK	1<<8
	SBLK	FRMSIZ+BINSIZ<<8 ; did it overflow
	BLEZ	CYC00		; branch if no
	ZAC			; yes. clamp at max
CYC00	ADLK	FRMSIZ+BINSIZ<<8	
	SACL	SYNCTR
;
; The following processing is done once for each bit. The mark/space
; differential signal is used to derive the regenerator output; the
; mark and space signals are used to derive signal and noise estimates.
; The ACCUM signal can be monitored on a DC-coupled scope connected to
; the AIO output using the d5 command. This is a very useful tuning
; indicator for use in performance assessment.
;
	LAC	MARK		; compute differential signal
	SUB	SPACE
	BIT	STATUS,RXREV	; is shift reversed
	BBZ	CYC12		; branch if no
	NEG			; yes. invert signal
CYC12	LARP	AR1		; AR1 is survivor table pointerl
	LRLK	AR1,SRVEND-2	; point to end of survivor table
	RPTK	SRVSIZ-2	; shift old samples
	DMOV	*-
	MAR	*+
	SACL	*		; insert new sample first
	LAC	MSPTR		; decrement mark memory pointer
	SUBK	1
	SBLK	MRKBGN		; is it below lower limit
	BGEZ	CYC15		; branch if no
	ADLK	SRVSIZ		; yes. wrap to upper limit
CYC15	ADLK	MRKBGN
	SACL	MSPTR
	LAR	AR1,MSPTR
	BIT	STATUS,RXREV	; is shift reversed
	BBNZ	CYC13		; branch if yes
	LAC	MARK		; insert mark in carrier memory
	SACL	*
	ADRK	SPCBGN-MRKBGN
	LAC	SPACE		; insert space in noise memory
	SACL	*
	B	CYC14
;
CYC13	LAC	SPACE		; insert space in carrier memory
	SACL	*
	ADRK	SPCBGN-MRKBGN
	LAC	MARK		; insert mark in noise memory
	SACL	*
CYC14
	LRLK	AR1,SRVEND-BINSIZ-1 ; point to next bit to send
	LAC	*
	SACL	ACCUM		; *** make visual signal for scope
;
; If output is in progress we don't want to go further in decode until
; the end of the character. Instead, we do a little dance to determine
; when to send the next bit and whether mark or space. Note that the
; output is clamped to mark-hold if the range is less than the
; threshold. This can happen only in sync mode when the transmitter
; pauses just before the autostart signal kicks out. 
;
	LAC	BITCTR		; is output in progess
	BZ	CYC11		; branch if no
	LAC	SRVCTR		; yes. is this strobe event
	BNZ	CYC10		; branch if no
	LACK	BINSIZ		; yes. reinitialize strobe counter
	SACL	SRVCTR
	LAC	CTHRESH		; is range above threshold
	SUB	POSMAX
	ADD	NEGMAX
	BGZ	CYC18A		; branch if no
	BIT	STATUS,PGATE	; yes. is pulse gate enabled
	BBZ	CYC18		; branch if no
	BIT	STATUS,AUTTST	; yes. is autostart up
	BBNZ	CYC18		; branch if yes
CYC18A	LALK	LIMIT		; no. force mark hold
	B	CYC19
;
CYC18	LRLK	AR1,SRVEND-BINSIZ-1 ; point to next bit to send
	LAC	POSMAX		; determine key polarity (+ = mark)
	ADD	NEGMAX
	SFR
	NEG
	ADD	*
CYC19	SACL	KEY
	LAC	BITCTR		; decrement bit counter
	SUBK	1
	SACL	BITCTR

	LALK	LIMIT/2		; *** make visual strobe for scope
	ADD	ACCUM
	SACL	ACCUM

CYC10	LAC	SRVCTR		; decrement strobe counter
	SUBK	1
	SACL	SRVCTR
	B	CYCEND
;
CYC11	LALK	LIMIT		; idle condition. force mark hold
	SACL	KEY
;
; Compute the maximum and minimum of all survivors. The loop is unrolled
; twice in order to reduce overhead; however, the frequent branches
; effectively stall the pipeline.
;
	LRLK	AR1,SRVBGN	; point to beginning of survivor table
	LRLK	AR0,SRVSIZ+SRVBGN ; point to end of survivor table
	LALK	08000h		; initialize for max search
	SACL	POSMAX
	LALK	03777h		; initialize for min search
	SACL	NEGMAX
CYC30	LAC	*		; is sample > max
	SUB	POSMAX
	BLEZ	CYC31		; branch if no
	ADD	POSMAX		; yes. update max
	SACL	POSMAX
	SAR	AR1,TMP
CYC31	LAC	*+		; is sample < min
	SUB	NEGMAX
	BGEZ	CYC32		; branch if no
	ADD	NEGMAX		; yes. update min
	SACL	NEGMAX
CYC32	LAC	*		; is sample > max
	SUB	POSMAX
	BLEZ	CYC33		; branch if no
	ADD	POSMAX		; yes. update max
	SACL	POSMAX
	SAR	AR1,TMP
CYC33	LAC	*+		; is sample < min
	SUB	NEGMAX
	BGEZ	CYC34		; branch if no
	ADD	NEGMAX		; yes. update min
	SACL	NEGMAX
CYC34	CMPR			; is search complete
	BBZ	CYC30		; branch if no
;
; Compute carrier and noise estimates. The carrier estimate uses a peak
; detector and sample-and-hold circuit; the noise estimate is an
; exponential average. Both estimates depend on detecting the time of
; the maximum differential signal. At that time the corresponding mark
; signal is almost certainly a carrier sample, while the corresponding
; space signal is almost certainly a noise sample. Note that older
; samples correspond to higher addresses in both the survivor table and
; mark/space memories.
;
	LAC	TMP		; compute mark memory offset
	SBLK	SRVBGN
	ADD	MSPTR
	SBLK	MRKBGN
	SBLK	SRVSIZ		; is it above upper limit
	BGEZ	CYC35		; branch if yes (wrap)
	ADLK	SRVSIZ		; no. restore
CYC35	ADLK	MRKBGN		; make into pointer
	SACL	TMP
	LAR	AR1,TMP
	LAC	*		; peak-detect carrier sample
	SACL	TMP
	SUB	CARDET		; is this positive peak
	BGEZ	CYC36		; branch if yes
	SACL	TMP		; no save the difference
	LAC	SPARE1		; is hold time exceeded
	SUBK	1
	BGZ	CYC37		; branch if yes
	LAC	TMP,CCONST	; no. carrier estimate is average
	ADDH	CARDET
	SACH	CARDET
	B	CYC38
;
CYC36	LAC	TMP		; carrier estimate is peak
	SACL	CARDET
	LALK	SRVSIZ*2	; hold for two character times
CYC37	SACL	SPARE1
CYC38	ADRK	SPCBGN-MRKBGN	; point to space register
	ZALR	NOISE		; average noise sample
	ADD	*,CCONST
	SUB	NOISE,CCONST
	SACH	NOISE
	LAC	CARDET		; is carrier up
	SUB	NOISE
	SUB	CTHRESH
	BGZ	CYC39		; branch if yes
	LALK	1<<CARTST	; no. clear carrier and sync bits
	ORK	1<<SYNTST
	CMPL
	AND	STATUS
	SACL	STATUS
	B	CYC39A
;
CYC39	LALK	1<<CARTST	; carrier up. set carrier bit
	OR	STATUS
	SACL	STATUS
CYC39A
;
; At this point the last 8 bits (64 samples) are in the survivor shift
; register. The oldest bit in the register is the stop bit of the
; previous character, the next oldest bit is the start bit of the
; current character and the youngest is the stop bit of that character.
; A survivor is defined as the vector of samples beginning in the oldest
; bit and extending through each bit in turn to the youngest bit. There
; are 8 survivors. one for each sample in the oldest bit. The problem
; here is to identify whether the oldest bit begins a character and
; which of the 8 survivors represents the most likely transmitted
; signal.
;
; Here we test for good start and stop bits and select the survivor of
; maximum distance. The maximum difference between the mark and space
; channels over the last 8 bits is called the signal range. The average
; of this range is the slice level. If the signal is above the slice
; level, it is a mark; otherwise, it is a space. The interval since the
; last start bit is called the frame time. In order for a character to
; be recognized, the frame time must be at least 7 bits. For signals
; with stop bits greater than one unit, we expect the frame time to be
; that much longer; however, in order to affect the frame time estimate,
; it must not be longer than 8 bits.
;
; The exponential average of the max survivor distance is called the
; autostart signal. The program operates in asynchronous mode until it
; locks on a character stream and the autostart signal has increased
; above a threshold (ATHRESH). If so, and if enabled, the start of a
; character is determined by the frame time (SYNDEL) and phase-lock
; loop.
;
	LAC	SYNCTR		; is sample time too early
	SBLK	FRMSIZ<<8
	BLZ	CYCEND		; branch if yes
	BIT	STATUS,SYNTST	; is this sync mode
	BBZ	CYC60		; branch if no
;
; Synchronous mode
;
; In a nominal 75-baud, 10 char/s system, the bit rate is 75 Hz and
; sample rate eight time times this, 600 Hz. The corresponding bit
; interval is 13.33 ms and sample interval 1.667 ms, respectively.
;
; The phase error amplitude is proportional to the difference between
; the maximum survivor strobe number (SRVCTR) and the midpoint of the
; bit (BINSIZ/2). This represents a phase detector (PD) gain of 8 V/(2
; pi radians). The VCO (SYNDEL) scaling is such that a change of 256
; results in a change of one 1.667-ms sample in one 0.1-s character
; time. This represents a VCO gain of 65.1e-6 (2 pi radians)/(V-s). The
; open-loop gain constant is then the product of the PD and VCO gains,
; 521e-6 1/s.
;
; At the smallest frequency resolution of 65.1e-6, the time to cross a
; 1.667-ms sample interval is about 25.6 s, or 256 characters. We use a
; maximum time constant of 256, which is implemented using a right shift
; of 8 (SCONST). Smaller time constants must be used during initial
; acquisition. In the optimum (Butterworth) design, the product of the
; time constant, gain constant and gain factor should be 1/2. To make
; this happen, the gain factor should be 1/(2 25.6 521e-6) = 37.5. We
; approximate this with a gain factor of 32, which is implemented using
; a left shift of 5 (PGAIN).
;
; The maximum range of the VCO is limited by the magnitude of the error
; signal. The phase error is normalized in q11 format. To achieve a gain
; factor of 32, the phase error must be shifted by 9 bits to the left.
; This gives 4 bits of precision (signed) in the frame time estimate.
; The maximum frequency deviation in this configuration is 1.053e-3 s/s.
; Therefore, we must rely on the asynchronous mode to reliably fill in
; the 4 bits remaining between the decimal point of the frame time and
; the high-order bit of the phase error.
; 
	LAC	SYNCTR		; sync mode. is sample time too early
	SUB	SYNDEL
	BLZ	CYCEND		; branch if yes
	SACL	SYNCTR		; no. update sync counter
	B	CYC68		; yes. continue to select survivor
;
; Asynchronous mode
;
; This little bit of paranoia insists that, (a) the signal gate is
; enabled (see below), (b) the range is above half threshold, (c) in the
; last eight bit times, the oldest bit (stop bit of the previous
; character) is a mark, (d) the next oldest bit (start bit of the
; current character) is a space, and (e) the youngest bit (stop bit of
; the current character) is a mark. These steps help to resist false
; starts due to impulse noise and fading and, especially, resist garbles
; due to noise.
;
CYC60	BIT	STATUS,CARTST	; async mode. is signal gate enabled
	BBZ	CYC67		; branch if no
	LAC	CTHRESH		; yes. is range above threshold
	SUB	POSMAX
	ADD	NEGMAX
	BGZ	CYC67		; branch if no
	LAC	POSMAX		; yes. compute slice level
	ADD	NEGMAX
	SFR
	SACL	STS
	LRLK	AR1,SRVBGN+BINSIZ-1 ;point to stop bit
	LAC	*
	SUB	STS		; is sample mark
	BLZ	CYC67		; branch if no
	ADRK	6*BINSIZ	; yes. point to start bit
	LAC	*
	SUB	STS		; is sample space
	BGZ	CYC67		; branch if no
	ADRK	BINSIZ		; yes. point to last stop bit
	LAC	*
	SUB	STS		; is sample mark
	BGEZ	CYC68		; branch if yes
;
; This is not the beginning of a character or an error has been found.
; This happens once per sample, which is much more often than once per
; character.
;
CYC67	ZALH	AUTOS		; downdate autostart
	SUB	AUTOS,ACONST-3
	SACH	AUTOS
	B	CYCEND
;
; We have found the beginning of a valid character. The outer loop
; computes the distance for each survivor ending in the stop bit.
; The inner loop computes the distance summed over all bits of the
; survivor. AR1 at this time points to the stop bit of the survivor. We
; already know that (a) the stop bit of the previous character is a
; mark, (b) the start bit of the current character is a space, and (c)
; the stop bit of the current character is a mark. In other words, we
; have already thrown away all survivors that don't meet these criteria.
;
CYC68	LRLK	AR1,SRVBGN	; initialize for survivor search
	LARK	AR2,BINSIZ-1
	LALK	08000h
	SACL	DIST
;
; We compute the distance for each survivor according to the Viterbi
; algorithm. Note that the distance here is the negative of the Viterbi
; distance, in the distance increases with increasing probability.
;
	LAC	POSMAX		; compute slice level
	ADD	NEGMAX
	SFR
	SACL	STS
CYC73	LARP	AR1		; AR1 is survivor table pointer
	LAC	*,16-2		; distance for stop bit
	SUB	NEGMAX,16-2
	SACH	TMP
	ADRK	BINSIZ
	LAC	STS,16-2	; distance for data bit 5
	SUB	*,16-2
	ABS
	ADDH	TMP
	SACH	TMP
	ADRK	BINSIZ
	LAC	STS,16-2	; distance for data bit 4
	SUB	*,16-2
	ABS
	ADDH	TMP
	SACH	TMP
	ADRK	BINSIZ
	LAC	STS,16-2	; distance for data bit 3
	SUB	*,16-2
	ABS
	ADDH	TMP
	SACH	TMP
	ADRK	BINSIZ
	LAC	STS,16-2	; distance for data bit 2
	SUB	*,16-2
	ABS
	ADDH	TMP
	SACH	TMP
	ADRK	BINSIZ
	LAC	STS,16-2	; distance for data bit 1
	SUB	*,16-2
	ABS
	ADDH	TMP
	ADRK	BINSIZ
	ADD	POSMAX,16-2	; distance for start bit
	SUB	*,16-2
	SUBH	DIST		; is this the max survivor
	BLEZ	CYC74		; brunch if no
	ADDH	DIST		; yes. breakfast only
	SACH	DIST
	SAR	AR1,TMP
	LALK	7*BINSIZ+SRVBGN-1 ; determine the dinner phase
	SUB	TMP
	SACL	SRVCTR
CYC74	SBRK	6*BINSIZ-1	; point to the next survivor
	LARP	AR2		; AR2 is repeat counter
	BANZ	CYC73
;
; The phase error is the difference between the survivor number and half
; the number of samples per bit. More precisely, that's 3.5 out of 8
; samples per bit (examine a diagram if you don't believe it.). The
; phase error is exponentially averaged to form the error signal used to
; drive the PLL. The exponential average of the distance is the
; autostart signal, which serves as the pulse gate signal. Do not fiddle
; the phase if the distance doesn't make the grade.
;
	ZALR	AUTOS		; average the autostart signal
	ADD	DIST,ACONST
	SUB	AUTOS,ACONST
	SACH	AUTOS
	LALK	BINSIZ-1<<11	; phase error at 3.5 samples
	SUB	SRVCTR,12
	SACL	STS
	ZALR	XPHASE		; yes. average the phase error signal
	ADD	STS,SCONST
	SUB	XPHASE,SCONST
	SACH	XPHASE
	BIT	STATUS,SYNTST	; is this sync mode
	BBZ	CYC88		; branch if no
;
; In synchronous mode we simply update the frame phase from the average
; phase error.
;
	ZALR	SYNCTR		; update frame phase
	ADD	XPHASE,PGAIN
	SACH	SYNCTR
	B	CYC80
;
; In async mode we measure the intercharacter interval and determine the
; frame time as the exponential average of the sync counter since the
; last start bit, adjusted by the survivor position. The intent is to
; provide a seamless transition to synchronous mode. In order for the
; frame time to be updated, the intercharacter interval must be less
; than 8 bits. The pulse gate is implemented here. If enabled and the
; autostart signal is not above the threshold, the output is clamped at
; mark.
;
CYC88	LAC	SYNCTR		; is sample time too late
	SBLK	FRMSIZ+BINSIZ<<8
	BGEZ	CYC89		; branch if yes
	ZALR	SYNDEL		; no. update frame frequency
	ADD	SYNCTR,TCONST
	SUB	SYNDEL,TCONST
	SACH	SYNDEL
CYC89	ZAC			; set up for next frame
	SACL	SYNCTR
CYC80	LACK	CHRSIZ-1	; arm the transmitter
	SACL	BITCTR

	LALK	2*LIMIT		; *** make visual scope sync pulse
	ADD	ACCUM
	SACL	ACCUM
;
; This is the end of sample processing. We do a little housekeeping and
; wait until the next sample. Note that the carrier signal (positive
; peak) less the noise (average) exceeds the threshold.
;
CYCEND	LAC	AUTOS		; is autostart up
	SUB	ATHRESH
	BGZ	CYC92		; branch if yes
	LALK	1<<AUTTST	; no. clear autostart and sync bits
	ORK	1<<SYNTST
	CMPL
	AND	STATUS
	SACL	STATUS
	B	CYC99
;
CYC92	LALK	1<<AUTTST	; autostart up. set autostart bit
	OR	STATUS
	SACL	STATUS
	BIT	STATUS,ENBTST	; is sync mode enabled
	BBZ	CYC99		; branch if no
	BIT	STATUS,CARTST	; yes. is carrier up
	BBZ	CYC99		; branch if no
	LALK	1<<SYNTST	; yes. set sync mode bit
	OR	STATUS
	SACL	STATUS
CYC99	RET			; declare victory

;
; The following two routines implement two different synchronous
; filter/detectors which process the input filtered and limited signal
; to produce baseband mark and space signals. Each routine is optimized
; for a different mode. It multiplies the input signal by sin/cos
; functions and processes the I and Q quadrature signals in four
; identical lowpass filters. It then computes the baseband signal as the
; square root of the sum of the squares of the I and Q channels.
;
; Matched filter
;
; This routine implements a filter/detector optimized for high
; selectivity. It uses synchronous matched filters tuned to the baud
; interval. A single interleaved shift register is used for the I and Q
; channels and mark and space frequencies.
;
; Variables
; AR1		channel structure pointer
; SIGNAL	input signal
; MFSIZE	size of matched filter (samples * 4)
; MFGAIN	matched filter scale
; GAIN		filter gain
; returns	filtered signal
;
MFILT	LARP	AR1		; AR1 is structure pointer
	LAC	*+		; (-increment)
	ADD	*		; (phase) did pointer underflow
	BGEZ	MFL01		; branch if no
	ADLK	PI2		; yes. adjust pointer
MFL01	SACL	*		; save and advance structure pointer
	LAC	*+,8		; scale for table offset
	SACH	STS
;
; I channel
;
	LAC	STS		; get sin argument
	ADLK	SINTAB
	TBLR	TMP
	LT	TMP		; sin * signal
	MPY	SIGNAL
	LTP	GAIN		; adjust gain (256+-)
	SACH	TMP
	MPY	TMP
	LTP	MFGAIN
	SACH	TMP,7
	LARP	AR0		; AR0 is MF delay line pointer
	LAR	AR0,MFPTR
	LAC	TMP		; add new value
	PUSH
	SUB	*		; subtract old value
	POPD	*+		; insert new value in mf
	SACL	TMP
	SAR	AR0,MFPTR
	LARP	AR1		; AR1 is structure pointer
	MPY	TMP		; scale for unity gain
	PAC
	ADDH	*+		; add to accumulator
	ADDS	*-
	SACH	*+
	SACL	*+
;
; Q channel
;
	LAC	STS		; get cos argument
	ADLK	30+SINTAB
	TBLR	TMP
	LT	TMP		; cos * signal
	MPY	SIGNAL
	LTP	GAIN		; adjust gain (256+-)
	SACH	TMP
	MPY	TMP
	LTP	MFGAIN
	SACH	TMP,7
	LARP	AR0		; AR0 is MF delay line pointer
	LAR	AR0,MFPTR
	LAC	TMP		; add new value
	PUSH
	SUB	*		; subtract old value
	POPD	*+		; insert new value in mf
	SACL	TMP
	SAR	AR0,MFPTR
	LARP	AR1		; AR1 is structure pointer
	MPY	TMP		; scale for unity gain
	PAC
	ADDH	*+		; add to accumulator
	ADDS	*-
	SACH	*+
	SACL	*+
	CALL	SQSQR		; square, sum and root
	RET

;
; Butterworth IIR filter
;
; This routine implements a filter/detector filter optimized for quasi-
; linear detector frequency response. It uses synchronous, second-order
; Butterworth IIR lowpass filters of two different bandwidths, one
; designed for Amateur 170-Hz shift, the other for commercial wide
; shifts. The bandwidth is determined by the coefficients loaded in
; block B0P. After detection, it processes the baseband signal in a
; matched filter tuned to the baud interval. The channel oscillators are
; spaced above and below the mark and space frequencies by 125 Hx, in
; order to improve linearity over the passband. The lowpass filters are
; designed with 6-dB rolloff equal to half the channel oscillator
; frequency spacing, so that the common-mode response is approximately
; uniform over the passband.
;
; Variables
; AR1		channel structure pointer
; AR2		IIR delay line pointer
; SIGNAL	input signal
; MFSIZE	size of matched filter (samples * 4)
; MFGAIN	matched filter scale
; GAIN		filter gain
; returns	filtered signal
;
IFILT	LARP	AR1		; AR1 is structure pointer
	LAC	*+		; (-increment)
	ADD	*		; (phase) did pointer underflow
	BGEZ	IFL01		; branch if no
	ADLK	PI2		; yes. adjust pointer
IFL01	SACL	*		; save and advance structure pointer
	LAC	*+,8		; scale for table offset
	SACH	STS
;
; I channel
;
	LAC	STS		; get sin argument
	ADLK	SINTAB
	TBLR	TMP
	LT	TMP		; sin * signal
	MPY	SIGNAL
	LTP	GAIN		; adjust gain (128/512+-)
	SACH	TMP
	MPY	TMP
	PAC
	LARP	AR2		; AR2 is IIR delay line pointer
	MAR	*+		; (AR1->z^-1)
	MPYK	0		; initialize sum and product registers
	MACD	LPFCOF,*-	; process a coefficients
	MACD	LPFCOF+1,*
	APAC
	SACH	*,5-1		; normalize
	MPYK	0		; initialize sum and product registers
	ZAC
	RPTK	3-1		; process b coefficients
	MAC	LPFCOF+3,*+
	APAC
	LARP	AR1		; AR1 is structure pointer
	SACH	*+
	SACL	*+
;
; Q channel
;
	LAC	STS		; get cos argument
	ADLK	30+SINTAB
	TBLR	TMP
	LT	TMP		; sin * signal
	MPY	SIGNAL
	LTP	GAIN		; adjust gain (128/512+-)
	SACH	TMP
	MPY	TMP
	PAC
	LARP	AR2		; AR2 is IIR delay line pointer
	MAR	*+		; (AR1->z^-1)
	MPYK	0		; initialize sum and product registers
	MACD	LPFCOF,*-	; process a coefficients
	MACD	LPFCOF+1,*
	APAC
	SACH	*,5-1		; normalize
	MPYK	0		; initialize sum and product registers
	ZAC
	RPTK	3-1		; process b coefficients
	MAC	LPFCOF+3,*+
	APAC
	LARP	AR1		; AR1 is structure pointer
	SACH	*+
	SACL	*+
	CALL	SQSQR		; square, sum and root
	SACL	TMP
;
; Process the baseband signal in a matched filter matched to the baud
; rate.
;
	LARP	AR0		; AR0 is MF delay line pointer
	LAR	AR0,MFPTR
	LAC	TMP
	PUSH
	SUB	*		; subtract old value
	POPD	*		; insert new value in mf
	SACL	TMP
	LACK	2		; increment filter pointer
	ADD	MFPTR
	SACL	MFPTR
	LARP	AR1		; AR1 is structure pointer
	LT	MFGAIN
	MPY	TMP		; scale for unity gain
	PAC
	ADRK	2		; add to accumulator
	ADDH	*+
	ADDS	*-
	SACH	*+
	SACL	*-
	LAC	*		; return high-order sum
	RET

;
; Subroutine to square, sum and root. AR1 points just past the
; accumulator
;
SQSQR	SBRK	4		; point to high order I channel sum
	MPYK	0		; initialize sum and product registers
	ZAC
	SQRA	*		; I^2
	ADRK	2
	SQRA	*		; Q^2
	ADRK	2
	APAC			; sum and save
	SACH	*+
	SACL	*-
;
; Find square root using Newton-Raphelson (3 iterations). Initial
; estimate for x assumes the signal is at the limit threshold.
;
	LALK	LIMIT		; initial guess
	SACL	TMP
	ZALH	*+		; (c / x) iteration 1
	ADDS	*-
	RPTK 	15-1
	SUBC	TMP
	SACL	STS
	LAC	STS		; (c / x + x)
	ADD	TMP
	SFR			; (1 / 2 * (c / x + x))
	SACL	TMP
	ZALH	*+		; (c / x) iteration 2
	ADDS	*-
	RPTK 	15-1
	SUBC	TMP
	SACL	STS
	LAC	STS		; (c / x + x)
	ADD	TMP
	SFR			; (1 / 2 * (c / x + x))
	SACL	TMP
	ZALH	*+		; (c / x) iteration 3
	ADDS	*-
	RPTK 	15-1
	SUBC	TMP
	SACL	STS
	LAC	STS		; (c / x + x)
	ADD	TMP
	SFR			; (1 / 2 * (c / x + x))
	RET

;
; Output signal processing. The AFSK output is synthesized here. Mark
; and space frequencies are determined according to mode and tables. For
; debugging purposes, provision is made to monitor various signals and
; state variables as selected by commands 0-7. Note that the output
; sinewave is attenuated about 10dB in order that A/B comparisons can be
; made conveniently using the attached TNC.
;
; Note: This is the only place where AR7 is incremented. Immediately
; after increment, AR7 is checked for overflow and wrapped.
;
; Variables
; AR7		interrupt buffer pointer
; KEY		output key line
; RANGE		baud interval counter radix
; OUTSW		display pointer
; OMARK		AFSK mark phase increment
; OSPACE	AFSK space phase increment
; OPHASE	AFSK output phase
;
OUTX	LAC	KEY		; is key line high
	BLZ	OUT11		; branch if no
	LAC	OMARK		; yes. set mark interval
	B	OUT12
;
OUT11	LAC	OSPACE		; set space interval
OUT12	ADD	OPHASE		; decrement cycle
	BGEZ	OUT10		; branch if no underflow
	ADLK	PI2		; underflow. add 2 * pi
OUT10	SACL	OPHASE
	LAC	OPHASE,8	; make table pointer
	SACH	UTMP1
	LALK	SINTAB
	ADD	UTMP1
	TBLR	STS		; convert to sin amplitude
	LALK	LIMIT		; scale for convenient A/B compare
	SACL	TMP
	LT	TMP
	MPY	STS
	PAC
	SACH	STS
	LARP	AR1		; AR1 is output sample pointer
	LAR	AR1,OUTSW	; fetch source signal in q15 format
	LAC	*
	ANDK    0FFFCh
	LARP	AR7		; AR7 is interrupt buffer pointer
	SACL	*+		; store output sample in q15 format
	LRLK	AR0,INTBUF+BUFLEN ; is pointer at buffer end
	CMPR	
	BBZ	OUT13		; branch if no
	LRLK	AR7,INTBUF	; yes. reset pointer
OUT13	RET

;
; The following routine implements a crude, table-driven command decoder
; which can be used to select various system parameters, such as baud
; rate, channel spacing, etc. It operates by sampling the UART receive-
; done bit. Someday this will be driven by interrupts; meanwhile, it
; should be called with interrupts disabled. Note that the UART hardware
; parameters are set up by the downloader.
;
; Variables
; UCHAR		character temp
; UPTR		pointer temp
; UADR		address temp
; UCMD		command table pointer
; UMOD		saved command table pointer
;
GETCH	DINT			; bolt the windows
	LDPK	B_PAGE
	LALK	UART_SEL_LSR	; select line status register
	SACL	UARTI
	OUT	UARTI,UART_CTRL
	IN	UARTI,UART_READ
	LALK	UART_LSR_RBF	; is input buffer full
	AND	UARTI
	BZ	GET14		; branch if no
	LALK	UART_SEL_RX	; yes. select input buffer
	SACL	UARTI
	OUT	UARTI,UART_CTRL 
	IN	UARTI,UART_READ	; input character
	LAC	UARTI
	LDPK	D_PAGE
	EINT			; allow fresh air
	ANDK	07fh		; unparity it
	SACL	UTMP1
GET13	LAC	UCMD		; initialize table pointer
	SACL	UPTR
GET10	LAC	UPTR		; get next routine address
	TBLR	UADR
	ADDK	1		; get argument
	TBLR	UTMP2
	ADDK	1		; get next command character
	TBLR	UCHAR
	ADDK	1
	SACL	UPTR
	LAC	UCHAR		; is this last entry
	BZ	GET11		; branch if yes
	SUB	UTMP1		; no. does character match
	BNZ	GET10		; branch if no
GET11	LAC	UADR		; yes. call routine
	LARP	AR1		; AR1 is utility pointer
	CALA
GET12	RET
;
GET14	LDPK	D_PAGE		; allow fresh air
	EINT
	RET
;
; Housekeeping staff
;
GETPSH	LAC	UCMD		; save command table pointer
	SACL	UMOD
	LAC	UTMP2
	SACL	UCMD
	RET
;
GETPOP	LAC	UMOD		; restore command table pointer
	SACL	UCMD
	B	GET13		; continue in command decode

;
; Following are command execution segments. For compactness, we assume
; the auxiliary register pointer is set at AR1.
;
; Restore to initial defaults
;
RESET	LARP	AR1		; (in case called directly)
	ZAC			; initialize things
	LRLK	AR1,B1D		; clear block B1 (0300-03ff)
	RPTK	100h-1
	SACL	*+
	LRLK	AR1,USRBGN	; clear external memory (0401-)
	LRLK	AR2,USREND-USRBGN
RST10	LARP	AR1
	SACL	*+
	LARP	AR2
	BANZ	RST10
	LARP	AR1
	LALK	ST00		; initialize command state machine
	SACL	UCMD
	LALK	BINSIZ/2+FRMSIZ<<8 ; preset frame interval
	SACL	SYNDEL
	LALK	ALIMIT		; preset autostart threshold
	SACL	ATHRESH
	LALK	CLIMIT		; preset carrier threshold
	SACL	CTHRESH
	LACK	5		; preset limiter gain
	SACL	LGAIN
	LALK	D_ADDR+STS	; preset monitoring point
	SACL	OUTSW
	LALK	BAUD3		; preset 75 baud
	CALL	CMDB1
	LALK	SHIFT7		; preset 575 Hz shift, IIR filter
	CALL	CMDM1
	RET
;
; Select mode, filter and shift m{1-4}
;
CMDM	LAC	UTMP2		; get initialization table pointer
CMDM1	LRLK	AR1,MSTRUCT	; set mark filter frequency
	TBLR	*
	ADDK	1
	LRLK	AR1,SSTRUCT	; set space filter frequency
	TBLR	*
	ADDK	1
	TBLR	TMP		; get mode bits
	ADDK	1		; set filter gain
	TBLR	GAIN
	ADDK	1
	TBLR	OMARK		; set output mark frequency
	ADDK	1
	TBLR	OSPACE		; set output space frequency
	LALK	IIRBIT+WIDBIT	; set mode bits
	CMPL
	AND	STATUS
	OR	TMP
	SACL	STATUS
	CNFD			; map B0 to data memory
	LRLK	AR1,B0D
	BIT	STATUS,WIDTST	; is wideband filter selected
	BBNZ	CMDM2		; branch if yes
	RPTK	FIRLEN-1	; no. use narrowband coefficients
	BLKP	BPF1,*+
	RPTK	6-1
	BLKP	IIR1,*+
	CNFP			; map B0 to program memory
	B	CMD55
;
CMDM2	RPTK	FIRLEN-1	; use wideband coefficients
	BLKP	BPF2,*+
	RPTK	6-1
	BLKP	IIR2,*+
	CNFP			; map B0 to program memory
	B	CMD55
;
; Select baud rate b{1-4}
;
CMDB	LAC	UTMP2		; get initialization table pointer
CMDB1	TBLR	RANGE		; AIO samples
	ADDK	1
	TBLR	MFSIZE		; MF size
	ADDK	1
	TBLR	MFGAIN		; MF gain
CMD55	ZAC
	LRLK	AR1,MSTRUCT+1	; clear mark filter structure
	RPTK	9-1
	SACL	*+
	LRLK	AR1,SSTRUCT+1	; clear space filter structure
	RPTK	9-1
	SACL	*+
	LRLK	AR1,DELIIR	; clear IIR delay line
	RPTK	12-1
	SACL	*
	LRLK	AR1,DELMFR	; clear MF delay line
	LAR	AR2,MFSIZE
CMD56	LARP	AR1		; AR1 is MF delay line pointer
	SACL	*+
	LARP	AR2		; AR2 is repeat counter
	BANZ	CMD56
	LARP	AR1		; (restore per convention)
	LALK	DELMFR		; leave MF pointer in handy place
	SACL	MFPTR
	LALK	MRKBGN
	SACL	MSPTR
	RET
;
; Select limiter gain l{0-5}
;
CMDL	LAC	UTMP2		; select limiter gain
	SACL	LGAIN
	RET
;
; Adjust space frequency t{+-}
;
CMDSP	LALK	-TUNEX		; t+ increase frequency
	B	CMD22

CMDSN	LALK	TUNEX		; t- reduce frequency
CMD22	LARP	AR1		; AR1 is structure pointer
	LRLK	AR1,SSTRUCT
	ADD	*
	SACL	*
	RET
;
; Adjust filter gain g{+-}
;
CMDGP	ZALR	GAIN		; g+ decrease gain
	SUB	GAIN,16-5
	SACH	GAIN
	RET
;
CMDGN	ZALR	GAIN		; g- increase gain
	ADD	GAIN,16-5
	SACH	GAIN
	RET
;
; Adjust autostart threshold a{+-}
;
CMDAP	ZALR	ATHRESH		; a+ decrease autostart synlimold
	SUB	ATHRESH,16-4
	SACH	ATHRESH
	RET
;
CMDAN	ZALR	ATHRESH		; a- increase autostart synlimold
	ADD	ATHRESH,16-4
	SACH	ATHRESH
	RET
;
; Adjust carrier threshold c{+-}
;
CMDCP	ZALR	CTHRESH		; c+ decrease carrier threshold
	SUB	CTHRESH,16-4
	SACH	CTHRESH
	RET
;
CMDCN	ZALR	CTHRESH		; c- increase carrier threshold
	ADD	CTHRESH,16-4
	SACH	CTHRESH
	RET
;
; Select display/output d{0-9}
;
CMDD	LAC	UTMP2		; select monitor point
	SACL	OUTSW
	RET
;
; Toggle bit l, p, s, x
;
CMDTOG	LALK	1<<SYNTST	; kill sync mode
	CMPL
	AND	STATUS
	XOR	UTMP2		; toggle status bit
	SACL	STATUS
	RET

;
; UART output watcher
;
; Variables
; UARTO		UART output buffer
; 
PUTCH	LALK	UART_SEL_LSR	; select line status register
	SACL	TMP
	OUT	TMP,UART_CTRL 
	IN	TMP,UART_READ	; is output buffer empty
	LACK	20h
	AND	TMP
	BZ	PUT10		; branch if no
	LALK	UART_LSR_THRE	; yes. select output buffer
	SACL	TMP
	OUT	TMP,UART_CTRL
	OUT	UARTO,UART_WRITE ; output character
PUT10	RET

;
; Subroutine to dazzle LEDs
;
LIGHTS	LALK	LED_203+LED_204	; LEDs 3-4 off
	CMPL
	LDPK	B_PAGE
	AND	PDC_BUF
	LDPK	D_PAGE
	SACL	UTMP1
	LALK	LED_205+LED_206+LED_207 ; LEDs 5-7 off
	CMPL
	LDPK	B_PAGE
	AND	TNC_BUF
	LDPK	D_PAGE
	SACL	UTMP2
	BIT	STATUS,CARTST	; is signal gate enabled
	BBZ	MAN02		; branch if no
	LALK	LED_203		; yes. LED 3 on
	OR	UTMP1
	SACL	UTMP1
MAN02	BIT	STATUS,SYNTST	; is this sync mode
	BBNZ	MAN03A		; branch if yes
	BIT	STATUS,AUTTST	; no. is pulse gate enabled
	BBZ	MAN03		; branch if no
	LAC	KEY		; yes. is output mark
	BLZ	MAN03		; branch if no
MAN03A	LALK	LED_204		; yes. LED 4 on
	OR	UTMP1
	SACL	UTMP1
MAN03	LALK	LIMIT		; is mark greater than threshold
	SUB	MARK
	BGZ	MAN00		; branch if no
	LALK	LED_205		; yes. LED 5 on
	OR	UTMP2
	SACL	UTMP2
MAN00	LALK	LIMIT		; is space greater than threshold
	SUB	SPACE
	BGZ	MAN01		; branch if no
	LALK	LED_206		; yes. LED 6 on
	OR	UTMP2
	SACL	UTMP2
MAN01	LAC	KEY		; is this mark
	BLZ	MAN04		; branch if no
	LALK	LED_207		; yes. LED 7 on
	OR	UTMP2
	SACL	UTMP2
MAN04	LAC	UTMP1
	LDPK	B_PAGE
	SACL	PDC_BUF
	OUT     PDC_BUF,RADIO_CTRL ; LEDs 3-4
	LDPK	D_PAGE
	LAC	UTMP2
	LDPK	B_PAGE
	SACL	TNC_BUF
	OUT	TNC_BUF,TNC_OUTPUT ; LEDs 5-7
	LDPK	D_PAGE
	RET

;
; Turn selected LED on
;
LEDON	LDPK	U_PAGE
	OR	PIO_BUF		; set LED bit
	SACL	PIO_BUF
	OUT	PIO_BUF,RADIO_GAIN
	LDPK	D_PAGE
	RET

;
; Turn selected LED off
;
LEDOFF	LDPK	U_PAGE
	CMPL			; clear LED bit
	AND	PIO_BUF
	SACL	PIO_BUF
	OUT	PIO_BUF,RADIO_GAIN
	LDPK	D_PAGE
	RET

;
; Wait loops
;
WAIT4	LARK	AR2,0FFh
	CALL	WAIT2
	CALL	WAIT2
	RET

WAIT2	LARK	AR2,0FFh
	CALL	WAIT1
	CALL	WAIT1
	RET

WAIT1	LARK	AR2,0FFh
WAIT_A	LARK	AR3,0FFh
WAIT_B	LARP	AR3
	BANZ	WAIT_B		; Loop through count
	LARP	AR2
	BANZ	WAIT_A		; Loop through count
	RET

	.MSFIRST

;
; Command tables
;
ST00	.WORD	GETPSH,ST40,'a'	; adjust autostart threshold
	.WORD	GETPSH,ST20,'b'	; select baud rate
	.WORD	GETPSH,ST41,'c'	; adjust carrier threshold
	.WORD	GETPSH,ST10,'d'	; select display/output
	.WORD	GETPSH,ST51,'g'	; adjust filter gain
	.WORD	GETPSH,ST30,'l'	; select limiter gain
	.WORD	GETPSH,ST70,'m'	; select mode, filter and shift
	.WORD	CMDTOG,1<<PGATE,'p' ; enable/disable pulse gate
	.WORD	RESET,0,'r'	; master reset
	.WORD	CMDTOG,1<<ENBTST,'s' ; enable/disable sync mode
	.WORD	GETPSH,ST50,'t'	; adjust space frequency
	.WORD	CMDTOG,1<<RXREV,'x' ; reverse shift
	.WORD	GET12,0,0	; default
;
; Select baud rate b<1-4>
;
ST20	.WORD	CMDB,BAUD1,'1'	; 45.45 baud
	.WORD	CMDB,BAUD2,'2'	; 50 baud
	.WORD	CMDB,BAUD3,'3'	; 75 baud
	.WORD	CMDB,BAUD4,'4'	; 100 baud
	.WORD	GETPOP,0,0	; default
;
; Select mode, filter and shift m{1-8}
;
ST70	.WORD	CMDM,SHIFT1,'1'	; MF filter, 170 Hz shift
	.WORD	CMDM,SHIFT2,'2'	; MF filter, 425 Hz shift
	.WORD	CMDM,SHIFT3,'3'	; MF filter, 575 Hz shift
	.WORD	CMDM,SHIFT4,'4'	; MF filter, 850 Hz shift
	.WORD	CMDM,SHIFT5,'5'	; IIR filter, 170 Hz shift
	.WORD	CMDM,SHIFT6,'6'	; IIR filter, 425 Hz shift
	.WORD	CMDM,SHIFT7,'7'	; IIR filter, 575 Hz shift
	.WORD	CMDM,SHIFT8,'8'	; IIR filter, 850 Hz shift
	.WORD	GETPOP,0,0	; default
;
; Select limiter gain
;
ST30	.WORD	CMDL,0,'0'	; 0dB
	.WORD	CMDL,1,'1'	; 6dB
	.WORD	CMDL,2,'2'	; 12dB
	.WORD	CMDL,3,'3'	; 18dB
	.WORD	CMDL,4,'4'	; 24dB
	.WORD	CMDL,5,'5'	; 30dB
;
; Select display (output) d<0-9>
;
ST10	.WORD	CMDD,D_ADDR+STS,'0' ; AFSK output
	.WORD	CMDD,D_ADDR+BUFI,'1' ; AIO input signal
	.WORD	CMDD,D_ADDR+SIGNAL,'2' ; filtered/limited input signal
	.WORD	CMDD,D_ADDR+MARK,'3' ; mark signal
	.WORD	CMDD,D_ADDR+SPACE,'4' ; space signal
	.WORD	CMDD,D_ADDR+ACCUM,'5' ; detector output
	.WORD	CMDD,D_ADDR+AUTOS,'6' ; autostart signal
	.WORD	CMDD,D_ADDR+XPHASE,'7' ; PLL phase signal
	.WORD	CMDD,D_ADDR+NOISE,'8' ; noise signal
	.WORD	CMDD,D_ADDR+CARDET,'9' ; carrier signal
	.WORD	GETPOP,0,0	; defaultt
;
; Adjust space frequency t{+-}
;
ST50	.WORD	CMDSP,0,'+'	; increase
	.WORD	CMDSP,0,'='	; increase (LC)
	.WORD	CMDSN,0,'-'	; decrease
	.WORD	GETPOP,0,0	; default
;
; Adjust filter gain g{+-}
;
ST51	.WORD	CMDGN,0,'+'	; increase
	.WORD	CMDGN,0,'='	; increase (LC)
	.WORD	CMDGP,0,'-'	; decrease
	.WORD	GETPOP,0,0	; default
;
; Adjust autostart threshold a{+-}
;
ST40	.WORD	CMDAN,0,'+'	; increase
	.WORD	CMDAN,0,'='	; increase (LC)
	.WORD	CMDAP,0,'-'	; decrease
	.WORD	GETPOP,0,0	; default
;
; Adjust carrier threshold c{+-}
;
ST41	.WORD	CMDCN,0,'+'	; increase
	.WORD	CMDCN,0,'='	; increase (LC)
	.WORD	CMDCP,0,'-'	; decrease
	.WORD	GETPOP,0,0	; default
;
; Initialization table. This is used to configure the AIO at initial
; startup.
;
INITAB	.WORD	3		; 0 load AIO configuration
	.WORD	CFGC		; 1
	.WORD	3		; 2 load TB register
	.WORD	CFGB		; 3
	.WORD	3		; 4 load TA register
	.WORD	CFGA		; 5
	.WORD	0		; 6 terminator

;
; Baud rate data tables. Note that 74.2 baud is rounded to 75 baud,
; since almost all RTTY today uses electronic UARTs. The first word is
; the number of AIO samples per bit sample, where there are 8 samples
; per bit. The next two words establish the matched filter size and
; gain, respectively. 
;
; 45.45 baud
;
BAUD1	.WORD	2750		; AIO samples (1000000 / (r * BINSIZ))
	.WORD	175*4		; MF size * 4 (SFREQ / r)
	.WORD	374		; MF gain (2^16 / size)
;
; 50 baud
;
BAUD2	.WORD	2500		; AIO samples (1000000 / (r * BINSIZ))
	.WORD	159*4		; MF size * 4 (SFREQ / r)
	.WORD	412		; MF gain (2^16 / size)
;
; 75 Baud
;
BAUD3	.WORD	1667		; AIO samples (1000000 / (r * BINSIZ))
	.WORD	106*4		; MF size * 4 (SFREQ / r)
	.WORD	618		; MF gain (2^16 / size)
;
; 100 baud
;
BAUD4	.WORD	1250		; AIO samples (1000000 / (r * BINSIZ))
	.WORD	79*4		; MF size * 4 (SFREQ / r)
	.WORD	830		; MF gain (2^16 / size)

;
; Tone frequency data tables. These are used to initialize variables for
; various combinations of mode, filter and spacing. The first two words
; control the input mark and space carrier oscillators, respectively.
; The third word controls various mode bits. The fourth word sets the
; filter gain n+m, where n is the calculated gain factor and m is a trim
; value. The fifth and sixth words control the output AFSK mark and
; space carrier oscillators. The MF filter gain is set empirically at
; 2295 Hz with the limiter switched off so that the signal LED (LED 4)
; comes on with the mark LED (LED 5). The IIR filter gains are set so
; the mark signal is the same as the MF value at each shift.
;
; MF filter, narrow shift (170 Hz)
;
SHIFT1	.WORD	-8869		; mark in (2295 * 120 * 256 / SFREQ)
	.WORD	-8212		; space in (2125 * 120 * 256 / SFREQ)
	.WORD	0		; narrow filter
	.WORD	256+8		; filter gain
	.WORD	-8212		; mark out (2125 * 120 * 256 / SFREQ)
	.WORD	-8869		; space out (2295 * 120 * 256 / SFREQ)
;
; MF filter, mid shift (425 Hz)
;
SHIFT2	.WORD	-8869		; mark in (2295 * 120 * 256 / SFREQ)
	.WORD	-7181		; space in (1870 * 120 * 256 / SFREQ)
	.WORD	WIDBIT		; wide filter
	.WORD	256+8		; filter gain
	.WORD	-4927		; mark out (1275 * 120 * 256 / SFREQ)
	.WORD	-8212		; space out (2125 * 120 * 256 / SFREQ)
;
; MF filter, mid shift (575 Hz)
;
SHIFT3	.WORD	-8869		; mark in (2295 * 120 * 256 / SFREQ)
	.WORD	-6605		; space in (1720 * 120 * 256 / SFREQ)
	.WORD	WIDBIT		; wide filter
	.WORD	256+8		; filter gain
	.WORD	-4927		; mark out (1275 * 120 * 256 / SFREQ)
	.WORD	-8212		; space out (2125 * 120 * 256 / SFREQ)
;
; MF filter, wide shift (850 Hz)
;
SHIFT4	.WORD	-8869		; mark in (2295 * 120 * 256 / SFREQ)
	.WORD	-5549		; space in (1445 * 120 * 256 / SFREQ)
	.WORD	WIDBIT		; wide filter
	.WORD	256+8		; filter gain
	.WORD	-4927		; mark out (1275 * 120 * 256 / SFREQ)
	.WORD	-8212		; space out (2125 * 120 * 256 / SFREQ)
;
; IIR filter, narrow shift (170 Hz)
;
SHIFT5	.WORD	-9352		; mark in (2420 * 120 * 256 / SFREQ)
	.WORD	-7729		; space in (2000 * 120 * 256 / SFREQ)
	.WORD	IIRBIT		; narrow filter
	.WORD	51		; filter gain
	.WORD	-8212		; mark out (2125 * 120 * 256 / SFREQ)
	.WORD	-8869		; space out (2295 * 120 * 256 / SFREQ)
;
; IIR filter, mid shift (425 Hz)
;
SHIFT6	.WORD	-9352		; mark in (2420 * 120 * 256 / SFREQ)
	.WORD	-6744		; space in (1745 * 120 * 256 / SFREQ)
	.WORD	IIRBIT+WIDBIT	; wide filter
	.WORD	180		; filter gain
	.WORD	-4927		; mark out (1275 * 120 * 256 / SFREQ)
	.WORD	-8212		; space out (2125 * 120 * 256 / SFREQ)
;
; IIR filter, mid shift (575 Hz)
;
SHIFT7	.WORD	-9352		; mark in (2420 * 120 * 256 / SFREQ)
	.WORD	-6164		; space in (1595 * 120 * 256 / SFREQ)
	.WORD	IIRBIT+WIDBIT	; wide filter
	.WORD	180		; filter gain
	.WORD	-4927		; mark out (1275 * 120 * 256 / SFREQ)
	.WORD	-8212		; space out (2125 * 120 * 256 / SFREQ)
;
; IIR filter, wide shift (850 Hz)
;
SHIFT8	.WORD	-9352		; mark in (2420 * 120 * 256 / SFREQ)
	.WORD	-5101		; space in (1320 * 120 * 256 / SFREQ)
	.WORD	IIRBIT+WIDBIT	; wide filter
	.WORD	180		; filter gain
	.WORD	-4927		; mark out (1275 * 120 * 256 / SFREQ)
	.WORD	-8212		; space out (2125 * 120 * 256 / SFREQ)

;
; The following filter coefficient tables were generated using Math
; Works Inc. Matlab Signal Processing Toolkit. The performance data were
; obtained directly from the rounded and truncated coefficients, in
; order to accurately model the roundoff error.
;
; IIR lpf coefficients for quasi-linear limiter/discriminator -
; Butterworth, order 2. Note these are centered 125 Hz wide of the
; working shift, in order to improve linearity; e.g., for 170-Hz shift,
; the mark filter is centered on 2420 Hz, the space filter on 2000 Hz.
;
; 170-Hz shift: (420 Hz spacing): 2000-2420 Hz for 2125-2295 shift
;
IIR1	.WORD	-1734		; a2 q10 150/188 Hz (170-Hz linear)
	.WORD	3755		; a1 q10
	.WORD	2048		; a0 q10
	.WORD	6711		; b0 q20
	.WORD	13421		; b1 q20
	.WORD	6711		; b2 q20
;
; 850-Hz shift (1100 Hz spacing): 1320-2420 Hz for 1445-2295 shift
;
IIR2	.WORD	-1278		; a2 q10 425/547 Hz (850-Hz linear)
	.WORD	3142		; a1 q10
	.WORD	2048		; a0 q10
	.WORD	11747		; b0 q18
	.WORD	23495		; b1 q18
	.WORD	11747		; b2 q18

;
; FIR bpf coefficients for 170-Hz shift - remez method, q15 format
; order 50, sample freq 8000, low pass 2125, high pass 2295, skirt 300,
; 6dB 1969-2453 Hz, 30dB 1844-2578 Hz, shape 1.516, min stop atten 36dB
;
BPF1	.WORD	  -201		; -25
	.WORD	   318		; -24
	.WORD	   251		; -23
	.WORD	  -230		; -22
	.WORD	   -81		; -21
	.WORD	   304		; -20
	.WORD	     4		; -19
	.WORD	  -175		; -18
	.WORD	    25		; -17
	.WORD	   -65		; -16
	.WORD	   159		; -15
	.WORD	   324		; -14
	.WORD	  -614		; -13
	.WORD	  -408		; -12
	.WORD	  1288		; -11
	.WORD	   135		; -10
	.WORD	 -1988		;  -9
	.WORD	   575		;  -8
	.WORD	  2440		;  -7
	.WORD	 -1630		;  -6
	.WORD	 -2397		;  -5
	.WORD	  2768		;  -4
	.WORD	  1763		;  -3
	.WORD	 -3646		;  -2
	.WORD	  -649		;  -1
	.WORD	  3975		;   0
	.WORD	  -649		;   1
	.WORD	 -3646		;   2
	.WORD	  1763		;   3
	.WORD	  2768		;   4
	.WORD	 -2397		;   5
	.WORD	 -1630		;   6
	.WORD	  2440		;   7
	.WORD	   575		;   8
	.WORD	 -1988		;   9
	.WORD	   135		;  10
	.WORD	  1288		;  11
	.WORD	  -408		;  12
	.WORD	  -614		;  13
	.WORD	   324		;  14
	.WORD	   159		;  15
	.WORD	   -65		;  16
	.WORD	    25		;  17
	.WORD	  -175		;  18
	.WORD	     4		;  19
	.WORD	   304		;  20
	.WORD	   -81		;  21
	.WORD	  -230		;  22
	.WORD	   251		;  23
	.WORD	   318		;  24
	.WORD	  -201		;  25
;
; FIR bpf coefficients for 850-Hz shift - remez method, q15 format
; order 50, sample freq 8000, low pass 1445, high pass 2295, skirt 300,
; 6dB 1297-2438 Hz, 30dB 1172-2578 Hz, shape 1.233, min stop atten 36dB
;
BPF2	.WORD	  -136		; -25
	.WORD	   274		; -24
	.WORD	   210		; -23
	.WORD	   -62		; -22
	.WORD	     2		; -21
	.WORD	  -103		; -20
	.WORD	  -402		; -19
	.WORD	   176		; -18
	.WORD	   718		; -17
	.WORD	   -54		; -16
	.WORD	  -458		; -15
	.WORD	     8		; -14
	.WORD	  -438		; -13
	.WORD	  -364		; -12
	.WORD	  1318		; -11
	.WORD	   916		; -10
	.WORD	 -1266		;  -9
	.WORD	  -743		;  -8
	.WORD	    36		;  -7
	.WORD	 -1107		;  -6
	.WORD	  1503		;  -5
	.WORD	  4523		;  -4
	.WORD	 -2010		;  -3
	.WORD	 -7970		;  -2
	.WORD	   924		;  -1
	.WORD	  9427		;   0
	.WORD	   924		;   1
	.WORD	 -7970		;   2
	.WORD	 -2010		;   3
	.WORD	  4523		;   4
	.WORD	  1503		;   5
	.WORD	 -1107		;   6
	.WORD	    36		;   7
	.WORD	  -743		;   8
	.WORD	 -1266		;   9
	.WORD	   916		;  10
	.WORD	  1318		;  11
	.WORD	  -364		;  12
	.WORD	  -438		;  13
	.WORD	     8		;  14
	.WORD	  -458		;  15
	.WORD	   -54		;  16
	.WORD	   718		;  17
	.WORD	   176		;  18
	.WORD	  -402		;  19
	.WORD	  -103		;  20
	.WORD	     2		;  21
	.WORD	   -62		;  22
	.WORD	   210		;  23
	.WORD	   274		;  24
	.WORD	  -136		;  25

;
; Sin table in 3-degree increments (450 degrees), q15 format
;
SINTAB	.WORD	     0		;   0
	.WORD	  1715		;   3
	.WORD	  3425		;   6
	.WORD	  5126		;   9
	.WORD	  6813		;  12
	.WORD	  8481		;  15
	.WORD	 10126		;  18
	.WORD	 11743		;  21
	.WORD	 13328		;  24
	.WORD	 14876		;  27
	.WORD	 16384		;  30
	.WORD	 17847		;  33
	.WORD	 19261		;  36
	.WORD	 20622		;  39
	.WORD	 21926		;  42
	.WORD	 23170		;  45
	.WORD	 24351		;  48
	.WORD	 25466		;  51
	.WORD	 26510		;  54
	.WORD	 27482		;  57
	.WORD	 28378		;  60
	.WORD	 29197		;  63
	.WORD	 29935		;  66
	.WORD	 30592		;  69
	.WORD	 31164		;  72
	.WORD	 31651		;  75
	.WORD	 32052		;  78
	.WORD	 32365		;  81
	.WORD	 32588		;  84
	.WORD	 32723		;  87
	.WORD	 32767		;  90
	.WORD	 32723		;  93
	.WORD	 32588		;  96
	.WORD	 32365		;  99
	.WORD	 32052		; 102
	.WORD	 31651		; 105
	.WORD	 31164		; 108
	.WORD	 30592		; 111
	.WORD	 29935		; 114
	.WORD	 29197		; 117
	.WORD	 28378		; 120
	.WORD	 27482		; 123
	.WORD	 26510		; 126
	.WORD	 25466		; 129
	.WORD	 24351		; 132
	.WORD	 23170		; 135
	.WORD	 21926		; 138
	.WORD	 20622		; 141
	.WORD	 19261		; 144
	.WORD	 17847		; 147
	.WORD	 16384		; 150
	.WORD	 14876		; 153
	.WORD	 13328		; 156
	.WORD	 11743		; 159
	.WORD	 10126		; 162
	.WORD	  8481		; 165
	.WORD	  6813		; 168
	.WORD	  5126		; 171
	.WORD	  3425		; 174
	.WORD	  1715		; 177
	.WORD	     0		; 180
	.WORD	 -1715		; 183
	.WORD	 -3425		; 186
	.WORD	 -5126		; 189
	.WORD	 -6813		; 192
	.WORD	 -8481		; 195
	.WORD	-10126		; 198
	.WORD	-11743		; 201
	.WORD	-13328		; 204
	.WORD	-14876		; 207
	.WORD	-16384		; 210
	.WORD	-17847		; 213
	.WORD	-19261		; 216
	.WORD	-20622		; 219
	.WORD	-21926		; 222
	.WORD	-23170		; 225
	.WORD	-24351		; 228
	.WORD	-25466		; 231
	.WORD	-26510		; 234
	.WORD	-27482		; 237
	.WORD	-28378		; 240
	.WORD	-29197		; 243
	.WORD	-29935		; 246
	.WORD	-30592		; 249
	.WORD	-31164		; 252
	.WORD	-31651		; 255
	.WORD	-32052		; 258
	.WORD	-32365		; 261
	.WORD	-32588		; 264
	.WORD	-32723		; 267
	.WORD	-32768		; 270
	.WORD	-32723		; 273
	.WORD	-32588		; 276
	.WORD	-32365		; 279
	.WORD	-32052		; 282
	.WORD	-31651		; 285
	.WORD	-31164		; 288
	.WORD	-30592		; 291
	.WORD	-29935		; 294
	.WORD	-29197		; 297
	.WORD	-28378		; 300
	.WORD	-27482		; 303
	.WORD	-26510		; 306
	.WORD	-25466		; 309
	.WORD	-24351		; 312
	.WORD	-23170		; 315
	.WORD	-21926		; 318
	.WORD	-20622		; 321
	.WORD	-19261		; 324
	.WORD	-17847		; 327
	.WORD	-16384		; 330
	.WORD	-14876		; 333
	.WORD	-13328		; 336
	.WORD	-11743		; 339
	.WORD	-10126		; 342
	.WORD	 -8481		; 345
	.WORD	 -6813		; 348
	.WORD	 -5126		; 351
	.WORD	 -3425		; 354
	.WORD	 -1715		; 357
	.WORD	    -0		; 360
	.WORD	  1715		; 363
	.WORD	  3425		; 366
	.WORD	  5126		; 369
	.WORD	  6813		; 372
	.WORD	  8481		; 375
	.WORD	 10126		; 378
	.WORD	 11743		; 381
	.WORD	 13328		; 384
	.WORD	 14876		; 387
	.WORD	 16384		; 390
	.WORD	 17847		; 393
	.WORD	 19261		; 396
	.WORD	 20622		; 399
	.WORD	 21926		; 402
	.WORD	 23170		; 405
	.WORD	 24351		; 408
	.WORD	 25466		; 411
	.WORD	 26510		; 414
	.WORD	 27482		; 417
	.WORD	 28378		; 420
	.WORD	 29197		; 423
	.WORD	 29935		; 426
	.WORD	 30592		; 429
	.WORD	 31164		; 430
	.WORD	 31651		; 435
	.WORD	 32052		; 438
	.WORD	 32365		; 441
	.WORD	 32588		; 444
	.WORD	 32723		; 447
	.WORD	 32767		; 450
;
	.END			; end of program
